@@ -97,7 +97,7 @@ pinfo_init (char **envp, int envc)
 
       myself.thisproc (NULL);
       myself->pgid = myself->sid = myself->pid;
-      myself->ctty = -1;
+      myself->ctty = CTTY_UNINITIALIZED;
       myself->uid = ILLEGAL_UID;
       myself->gid = ILLEGAL_GID;
       environ_init (NULL, 0);	/* call after myself has been set up */
@@ -118,7 +118,7 @@ pinfo_init (char **envp, int envc)
 DWORD
 pinfo::status_exit (DWORD x)
 {
-  switch (x)
+  switch ((NTSTATUS) x)
     {
     case STATUS_DLL_NOT_FOUND:
       {
@@ -212,7 +212,7 @@ pinfo::exit (DWORD n)
     maybe_set_exit_code_from_windows ();	/* may block */
   exit_state = ES_FINAL;
 
-  if (myself->ctty > 0 && !iscons_dev (myself->ctty))
+  if (CTTY_IS_VALID (myself->ctty) && !iscons_dev (myself->ctty))
     {
       lock_ttys here;
       tty *t = cygwin_shared->tty[device::minor(myself->ctty)];
@@ -466,6 +466,8 @@ pinfo::init (pid_t n, DWORD flag, HANDLE h0)
       h = h0;
       _pinfo_release ();
     }
+  if (shloc == SH_MYSELF)
+    cygheap->shared_regions.myself_shared_addr = procinfo;
 }
 
 void
@@ -512,7 +514,7 @@ pinfo::pinfo (HANDLE parent, pinfo_minimal& from, pid_t pid):
 const char *
 _pinfo::_ctty (char *buf)
 {
-  if (ctty <= 0)
+  if (!CTTY_IS_VALID (ctty))
     strcpy (buf, "no ctty");
   else
     {
@@ -528,26 +530,32 @@ _pinfo::set_ctty (fhandler_termios *fh, int flags)
 {
   tty_min& tc = *fh->tc ();
   debug_printf ("old %s, ctty device number %y, tc.ntty device number %y flags & O_NOCTTY %y", __ctty (), ctty, tc.ntty, flags & O_NOCTTY);
-  if (fh && (ctty <= 0 || ctty == tc.ntty) && !(flags & O_NOCTTY))
+  if (fh && (!CTTY_IS_VALID (ctty) || ctty == tc.ntty) && !(flags & O_NOCTTY))
     {
-      ctty = tc.ntty;
-      if (cygheap->ctty != fh->archetype)
+      if (tc.getsid () && tc.getsid () != sid && ctty == CTTY_RELEASED)
+	; /* Do nothing if another session is associated with the TTY. */
+      else
 	{
-	  debug_printf ("cygheap->ctty %p, archetype %p", cygheap->ctty, fh->archetype);
-	  if (!cygheap->ctty)
-	    syscall_printf ("ctty was NULL");
-	  else
+	  ctty = tc.ntty;
+	  if (cygheap->ctty != fh->archetype)
 	    {
-	      syscall_printf ("ctty %p, usecount %d", cygheap->ctty,
-			      cygheap->ctty->archetype_usecount (0));
-	      cygheap->ctty->close ();
-	    }
-	  cygheap->ctty = (fhandler_termios *) fh->archetype;
-	  if (cygheap->ctty)
-	    {
-	      fh->archetype_usecount (1);
-	      /* guard ctty fh */
-	      report_tty_counts (cygheap->ctty, "ctty", "");
+	      debug_printf ("cygheap->ctty %p, archetype %p",
+			    cygheap->ctty, fh->archetype);
+	      if (!cygheap->ctty)
+		syscall_printf ("ctty was NULL");
+	      else
+		{
+		  syscall_printf ("ctty %p, usecount %d", cygheap->ctty,
+				  cygheap->ctty->archetype_usecount (0));
+		  cygheap->ctty->close ();
+		}
+	      cygheap->ctty = (fhandler_termios *) fh->archetype;
+	      if (cygheap->ctty)
+		{
+		  fh->archetype_usecount (1);
+		  /* guard ctty fh */
+		  report_tty_counts (cygheap->ctty, "ctty", "");
+		}
 	    }
 	}
 
@@ -570,13 +578,14 @@ _pinfo::set_ctty (fhandler_termios *fh, int flags)
 	 an obvious bug surfaces. */
       if (sid == pid && !tc.getsid ())
 	tc.setsid (sid);
-      sid = tc.getsid ();
+      if (CTTY_IS_VALID (ctty))
+	sid = tc.getsid ();
       /* See above */
       if ((!tc.getpgid () || being_debugged ()) && pgid == pid)
 	tc.setpgid (pgid);
     }
   debug_printf ("cygheap->ctty now %p, archetype %p", cygheap->ctty, fh ? fh->archetype : NULL);
-  return ctty > 0;
+  return CTTY_IS_VALID (ctty);
 }
 
 /* Test to determine if a process really exists and is processing signals.
@@ -597,6 +606,19 @@ _pinfo::alive ()
   return !!h;
 }
 
+static commune_result
+commune_process_siginfo ()
+{
+  commune_result res = { 0 };
+
+  res.pnd = sig_send (myself, __SIGPENDINGALL, NULL);
+  res.blk = cygheap->compute_sigblkmask ();
+  for (int sig = 1; sig < NSIG; ++sig)
+    if (global_sigs[sig].sa_handler == SIG_IGN)
+      res.ign |= SIGTOMASK (sig);
+  return res;
+}
+
 DWORD
 commune_process (void *arg)
 {
@@ -673,13 +695,7 @@ commune_process (void *arg)
     case PICOM_SIGINFO:
       {
 	sigproc_printf ("processing PICOM_SIGINFO");
-	commune_result cr;
-	sigpending (&cr.pnd);
-	cr.pnd = sig_send (myself, __SIGPENDINGALL, NULL);
-	cr.blk = cygheap->compute_sigblkmask ();
-	for (int sig = 1; sig < NSIG; ++sig)
-	  if (global_sigs[sig].sa_handler == SIG_IGN)
-	    cr.ign |= SIGTOMASK (sig);
+	commune_result cr = commune_process_siginfo ();
 	if (!WritePipeOverlapped (tothem, &cr, sizeof cr, &nr, 1000L))
 	  sigproc_printf ("WritePipeOverlapped siginfo failed, %E");
 	break;
@@ -1020,24 +1036,17 @@ _pinfo::root (size_t& n)
 int
 _pinfo::siginfo (sigset_t &pnd, sigset_t &blk, sigset_t &ign)
 {
+  commune_result cr;
+
   if (!pid)
     return -1;
   if (pid != myself->pid && !ISSTATE (this, PID_NOTCYGWIN))
-    {
-      commune_result cr = commune_request (PICOM_SIGINFO);
-      pnd = cr.pnd;
-      blk = cr.blk;
-      ign = cr.ign;
-    }
+    cr = commune_request (PICOM_SIGINFO);
   else
-    {
-      pnd = sig_send (myself, __SIGPENDINGALL, NULL);
-      blk = cygheap->compute_sigblkmask ();
-      ign = 0;
-      for (int sig = 1; sig < NSIG; ++sig)
-	if (global_sigs[sig].sa_handler == SIG_IGN)
-	  ign |= SIGTOMASK (sig);
-    }
+    cr = commune_process_siginfo ();
+  pnd = cr.pnd;
+  blk = cr.blk;
+  ign = cr.ign;
   return -1;
 }
 
